#!/usr/bin/env python3

"""Module containing the PreparePDB class and the command line interface."""
import argparse
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class PreparePDB(BiobbObject):
    """
    | biobb_cmip PreparePDB
    | Class to add CMIP charges and atom types.
    | Add CMIP charges and atom types to a PDB structure using `biobb_structure_checking <https://anaconda.org/bioconda/biobb_structure_checking>`_.

    Args:
        input_pdb_path (str): Input PDB file path. File type: input. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/data/cmip/1aki.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_cmip_pdb_path (str): Output PDB file path. File type: output. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/egfr_cmip.pdb>`_. Accepted formats: pdb (edam:format_1476).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **remove_water** (*bool*) - (True) Remove Water molecules.
            * **add_hydrogen** (*bool*) - (True) Add Hydrogen atoms to the structure.
            * **keep_hydrogen** (*bool*) - (False) If **add_hydrogen** is True. All hydrogen atoms will be removed before adding the new ones unless this option is set True.
            * **fix_sidechains** (*bool*) - (True) Complete side chains (heavy atoms, protein only).
            * **fix_backbone_atoms** (*bool*) - (True) Add missing O, OXT backbone atoms.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.


    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_cmip.cmip.prepare_pdb import prepare_pdb
            prop = { 'restart': False }
            prepare_pdb(input_pdb_path='/path/to/myStructure.pdb',
                        output_cmip_pdb_path='/path/to/newStructure.pdb',
                        properties=prop)

    Info:
        * wrapped_software:
            * name: CMIP cmip
            * version: 2.7.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(self, input_pdb_path: str, output_cmip_pdb_path: str, properties: dict = None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_pdb_path": input_pdb_path},
            "out": {"output_cmip_pdb_path": output_cmip_pdb_path}
        }

        # Properties specific for BB
        self.check_structure_path = properties.get('check_structure_path', 'check_structure')
        self.remove_water = properties.get('remove_water', True)
        self.keep_hydrogen = properties.get('keep_hydrogen', False)
        self.fix_sidechains = properties.get('fix_sidechains', True)
        self.fix_backbone_atoms = properties.get('fix_backbone_atoms', True)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`PreparePDB <cmip.prepare_pdb.PreparePDB>` object."""

        # Setup Biobb
        if self.check_restart(): return 0
        self.stage_files()

        command_list = ""
        if self.remove_water:
            command_list += 'water --remove yes; '

        if self.fix_backbone_atoms:
            command_list += 'backbone --add_caps none; '

        if self.fix_sidechains:
            command_list += 'fixside --fix All; '

        command_list += 'add_hydrogen --add_mode auto '
        if self.keep_hydrogen:
            command_list += ' -keep_h '
        command_list += '--add_charges CMIP'

        self.cmd = [self.check_structure_path,
                    '-v',
                    '-i', self.stage_io_dict["in"]["input_pdb_path"],
                    '-o', self.stage_io_dict["out"]["output_cmip_pdb_path"],
                    '--output_format', 'cmip',
                    '--non_interactive',
                    'command_list',
                    '--list', "'"+command_list+"'"]

        # Run Biobb block
        self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([self.stage_io_dict.get("unique_dir")])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def prepare_pdb(input_pdb_path: str, output_cmip_pdb_path: str, properties: dict = None, **kwargs) -> int:
    """Create :class:`PreparePDB <cmip.prepare_pdb.PreparePDB>` class and
    execute the :meth:`launch() <cmip.prepare_pdb.PreparePDB.launch>` method."""
    return PreparePDB(input_pdb_path=input_pdb_path,
                      output_cmip_pdb_path=output_cmip_pdb_path,
                      properties=properties, **kwargs).launch()


def main():
    parser = argparse.ArgumentParser(description="Model the missing atoms in the backbone of a PDB structure.",
                                     formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('-c', '--config', required=False, help="This file can be a YAML file, JSON file or JSON string")

    # Specific args of each building block
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('-i', '--input_pdb_path', required=True, help="Input PDB file name")
    required_args.add_argument('-o', '--output_cmip_pdb_path', required=True, help="Output PDB file name")

    args = parser.parse_args()
    config = args.config if args.config else None
    properties = settings.ConfReader(config=config).get_prop_dic()

    # Specific call of each building block
    prepare_pdb(input_pdb_path=args.input_pdb_path,
                output_cmip_pdb_path=args.output_cmip_pdb_path,
                properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the IgnoreResidues class and the command line interface."""
import argparse
import shutil
from biobb_cmip.cmip.common import mark_residues
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger
import biobb_common.tools.file_utils as fu


class IgnoreResidues(BiobbObject):
    """
    | biobb_cmip IgnoreResidues
    | Class to ignore residues in CMIP potential calculations.
    | Mark residues which will be ignored in the CMIP potential calculations except for dielectric definition.

    Args:
        input_cmip_pdb_path (str): Input PDB file path. File type: input. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/data/cmip/input_ignore_res.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_cmip_pdb_path (str): Output PDB file path. File type: output. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/ignore_res_gln3.pdb>`_. Accepted formats: pdb (edam:format_1476).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **residue_list** (*str*) - (None) Residue list in the format "Chain:Resnum" (no spaces between the elements) separated by commas. If no chain is provided all the residues in the pdb file will be market. ie: "A:3".
            * **ignore_all** (*bool*) - (False) Mark all the residues in the PDB file.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.


    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_cmip.cmip.ignore_residues import ignore_residues
            prop = { 'residue_list': "A:3" }
            ignore_residues(input_cmip_pdb_path='/path/to/myStructure.pdb',
                            output_cmip_pdb_path='/path/to/newStructure.pdb',
                            properties=prop)

    Info:
        * wrapped_software:
            * name: CMIP cmip
            * version: 2.7.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(self, input_cmip_pdb_path: str, output_cmip_pdb_path: str, properties: dict = None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_cmip_pdb_path": input_cmip_pdb_path},
            "out": {"output_cmip_pdb_path": output_cmip_pdb_path}
        }

        # Properties specific for BB
        self.residue_list = properties.get('residue_list', None)
        self.ignore_all = properties.get('ignore_all', False)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`IgnoreResidues <cmip.ignore_residues.IgnoreResidues>` object."""

        # Setup Biobb
        if self.check_restart(): return 0

        if not self.ignore_all and not self.residue_list:
            fu.log(f"Residue_list is empty and ignore_all is false nothing will be done.", self.out_log, self.global_log)
            shutil.copyfile(self.io_dict["in"]["input_cmip_pdb_path"], self.io_dict["out"]["output_cmip_pdb_path"])
            return self.return_code

        if self.ignore_all:
            self.residue_list = None

        if self.residue_list:
            if not isinstance(self.residue_list, list):
                self.residue_list = str(self.residue_list).split(",")
            for i in range(len(self.residue_list)):
                residue_code_list = str(self.residue_list[i]).split(":")
                if len(residue_code_list) < 2:
                    resnum = residue_code_list[0]
                    chain = ''
                else:
                    chain, resnum = residue_code_list
                self.residue_list[i] = chain.strip().upper()+":"+str(resnum).strip()

        mark_residues(residue_list=self.residue_list, input_cmip_pdb_path=self.io_dict["in"]["input_cmip_pdb_path"], output_cmip_pdb_path=self.io_dict["out"]["output_cmip_pdb_path"], out_log=self.out_log, global_log=self.global_log)

        # remove temporary
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def ignore_residues(input_cmip_pdb_path: str, output_cmip_pdb_path: str, properties: dict = None, **kwargs) -> int:
    """Create :class:`IgnoreResidues <cmip.ignore_residues.IgnoreResidues>` class and
    execute the :meth:`launch() <cmip.ignore_residues.IgnoreResidues.launch>` method."""
    return IgnoreResidues(input_cmip_pdb_path=input_cmip_pdb_path,
                          output_cmip_pdb_path=output_cmip_pdb_path,
                          properties=properties, **kwargs).launch()


def main():
    parser = argparse.ArgumentParser(description="Mark residues which charges will be ignored in the CMIP potential calculations.",
                                     formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('-c', '--config', required=False, help="This file can be a YAML file, JSON file or JSON string")

    # Specific args of each building block
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('-i', '--input_cmip_pdb_path', required=True, help="Input PDB file name")
    required_args.add_argument('-o', '--output_cmip_pdb_path', required=True, help="Output PDB file name")

    args = parser.parse_args()
    config = args.config if args.config else None
    properties = settings.ConfReader(config=config).get_prop_dic()

    # Specific call of each building block
    ignore_residues(input_cmip_pdb_path=args.input_cmip_pdb_path,
                    output_cmip_pdb_path=args.output_cmip_pdb_path,
                    properties=properties)


if __name__ == '__main__':
    main()



""" Common functions for package biobb_cmip.cmip """
import os
import re
import json
from pathlib import Path
from typing import List, Dict, Mapping, Union, Tuple, Sequence
import MDAnalysis as mda
from MDAnalysis.topology.guessers import guess_atom_element
import uuid
import logging
import biobb_common.tools.file_utils as fu


def get_grid(cmip_log_path: Union[str, Path], external: bool = False) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
    with open(cmip_log_path) as log_file:
        first_line = log_file.readline().strip()
    if first_line.startswith("titleParam"):
        return _get_grid_from_key_value(cmip_log_path, external)
    elif first_line.startswith("{"):
        return _get_grid_from_box_file(cmip_log_path)
    return _get_grid_from_text(cmip_log_path, external)


def _get_grid_from_box_file(cmip_box_path: Union[str, Path]) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
    with open(cmip_box_path) as json_file:
        grid_dict = json.load(json_file)
    origin = grid_dict['origin']['x'], grid_dict['origin']['y'], grid_dict['origin']['z']
    size = grid_dict['size']['x'], grid_dict['size']['y'], grid_dict['size']['z']
    return origin, size, grid_dict['params']


def _get_grid_from_text(cmip_log_path: Union[str, Path], external: bool = False) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
    origin = None
    size = None
    grid_params = {"CEN": None, "DIM": None, "INT": None}
    grid_locators_list = ["AUTOMATIC GRID", "MANUAL GRID"]
    if external:
        grid_locators_list = ["AUTOMATIC OUTER GRID", "MANUAL OUTER GRID"]

    with open(cmip_log_path) as log_file:
        inside_automatic_grid = False
        for line in log_file:
            if line.strip() in grid_locators_list:
                inside_automatic_grid = True
            if inside_automatic_grid:
                origin_match = re.match(r"Grid origin:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if origin_match:
                    origin = float(origin_match.group(1)), float(origin_match.group(2)), float(origin_match.group(3))
                size_match = re.match(r"Grid Size:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if size_match:
                    size = float(size_match.group(1)), float(size_match.group(2)), float(size_match.group(3))
                int_match = re.match(r"Grid units:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if int_match:
                    grid_params['INT'] = float(int_match.group(1)), float(int_match.group(2)), float(int_match.group(3))
                cen_match = re.match(r"Grid center:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))",line.strip())
                if cen_match:
                    grid_params['CEN'] = float(cen_match.group(1)), float(cen_match.group(2)), float(cen_match.group(3))
                dim_match = re.match(r"Grid density:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if dim_match:
                    grid_params['DIM'] = int(dim_match.group(1)), int(dim_match.group(2)), int(dim_match.group(3))
                if origin and size and grid_params['INT'] and grid_params['CEN'] and grid_params['DIM']:
                    break
    return origin, size, grid_params


def _get_grid_from_key_value(cmip_log_path: Union[str, Path], external: bool = False) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
    origin = None
    size = None
    grid_params = {"CEN": None, "DIM": None, "INT": None}
    grid_locators_list = ["AUTOMATIC GRID", "MANUAL GRID"]
    if external:
        grid_locators_list = ["AUTOMATIC OUTER GRID", "MANUAL OUTER GRID"]

    with open(cmip_log_path) as log_file:
        inside_automatic_grid = False
        for line in log_file:
            if line.strip() in grid_locators_list:
                inside_automatic_grid = True
            if inside_automatic_grid:
                origin_match = re.match(r"origin=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if origin_match:
                    origin = float(origin_match.group(1)), float(origin_match.group(2)), float(origin_match.group(3))
                size_match = re.match(r"size=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if size_match:
                    size = float(size_match.group(1)), float(size_match.group(2)), float(size_match.group(3))
                int_match = re.match(r"spacing=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if int_match:
                    grid_params['INT'] = float(int_match.group(1)), float(int_match.group(2)), float(int_match.group(3))
                cen_match = re.match(r"center=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))",line.strip())
                if cen_match:
                    grid_params['CEN'] = float(cen_match.group(1)), float(cen_match.group(2)), float(cen_match.group(3))
                dim_match = re.match(r"dim=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
                if dim_match:
                    grid_params['DIM'] = int(dim_match.group(1)), int(dim_match.group(2)), int(dim_match.group(3))
                if origin and size and grid_params['INT'] and grid_params['CEN'] and grid_params['DIM']:
                    break
    return origin, size, grid_params



def create_unique_file_path(parent_dir: Union[str, Path] = None, extension: str = None) -> str:
    if not parent_dir:
        parent_dir = Path.cwd
    if not extension:
        extension = ''
    while True:
        name = str(uuid.uuid4())+extension
        file_path = Path.joinpath(Path(parent_dir).resolve(), name)
        if not file_path.exists():
            return str(file_path)


def write_cmip_pdb(input_pdb_path, output_pdb_path, charges_list, elements_list):
    with open(input_pdb_path) as inPDB, open(output_pdb_path, 'w') as outPDB:
        index = 0
        for line in inPDB:
            line = line.rstrip()
            if not re.match('^ATOM', line) and not re.match('^HETATM', line):
                continue
            outPDB.write("{}{:8.4f}  {}\n".format(line[:54], charges_list[index], elements_list[index]))
            index += 1


def get_topology_cmip_elements_canonical(input_topology_filename: str) -> List:
    """
    This function also accepts pdb files
    Args:
        input_topology_filename:

    Returns:

    """
    # Remove forcefield itp references from top file.
    if input_topology_filename.lower().endswith('.top'):
        with open(input_topology_filename) as tf:
            top_lines = tf.readlines()
        top_file = create_unique_file_path(parent_dir=Path(input_topology_filename).parent.resolve(), extension='.top')
        with open(top_file, 'w') as nt:
            for line in top_lines:
                if re.search(r"\.ff.*\.itp", line):
                    continue
                nt.write(line)
        u = mda.Universe(top_file, topology_format="ITP")
        os.unlink(top_file)
    else:
        u = mda.Universe(input_topology_filename)
    # mda_charges = [round(val, 4) for val in u.atoms.charges]
    # mda_atom_types = list(guess_types(u.atoms.names))
    mda_atom_types = []
    for atom in u.atoms:
        atom_element = guess_atom_element(atom.name)
        if atom_element == 'H':
            bonded_atom_element = guess_atom_element(atom.bonded_atoms[0].name)
            if bonded_atom_element == 'O':
                atom_element = 'HO'
            elif bonded_atom_element in ['N', 'S']:
                atom_element = 'HN'
        mda_atom_types.append(atom_element)
    return mda_atom_types


def get_topology_charges(input_topology_filename: str) -> List:
    """ Given a topology which includes charges
    Extract those charges and save them in a list to be returned
    Supported formats (tested): prmtop, top, psf
    """
    # Remove forcefield itp references from top file.
    if input_topology_filename.lower().endswith('.top'):
        with open(input_topology_filename) as tf:
            top_lines = tf.readlines()
        top_file = create_unique_file_path(parent_dir=Path(input_topology_filename).parent.resolve(), extension='.top')

        with open(top_file, 'w') as nt:
            for line in top_lines:
                if re.search(r"\.ff.*\.itp", line):
                    continue
                nt.write(line)
        u = mda.Universe(top_file, topology_format="ITP")
        os.unlink(top_file)
    else:
        u = mda.Universe(input_topology_filename)
    return [round(val, 4) for val in u.atoms.charges]


class Residue:
    def __init__(self, data):
        self.id = data[0]+':'+data[1]
        self.atType = data[2]
        self.charg = float(data[3])


class ResiduesDataLib:
    def __init__(self, fname):
        self.RData = {}
        with open(fname) as fh:
            for line in fh:
                if line[0] == '#':
                    continue
                data = line.split()
                r = Residue(data)
                self.RData[r.id]=r
            self.nres = len(self.RData)

    def getParams(self, resid, atid):
        if resid+':'+atid in self.RData:
            return self.RData[resid+':'+atid]
        else:
            print("WARNING: atom not found in library (",resid+':'+atid,')')
            return {}


def get_pdb_charges(input_pdb_filename: str, residue_library_path: str = None) -> List:
    if not residue_library_path:
        residue_library_path = str(Path(__file__).parent.joinpath("dat", "aa.lib").resolve())

    aaLib = ResiduesDataLib(residue_library_path)
    print("{} residue/atom pairs loaded from {}".format(aaLib.nres, residue_library_path))

    with open(input_pdb_filename) as inPDB:
        charges_list = []
        residue_num = None
        for line in inPDB:
            line = line.rstrip()
            if not re.match('^ATOM', line) and not re.match('^HETATM', line):
                continue

            nomat = line[12:16]
            if re.match('^[1-9]', nomat):
                nomat = nomat[1:4] + nomat[:1]
            nomat = nomat.replace(' ', '')
            nomr = line[17:21].replace(' ', '')
            # WARNING: Temporal totally uninformed assumption by PA
            if nomr == "HIS":
                nomr = "HID"
                if residue_num != line[23:27]:
                    print(f"WARNING replacing HIS:{line[23:27]} by HID")
                    residue_num = line[23:27]
            # Thats not correct REVIEW this should be done for all the atoms in the residue
            # not just the oxigen
            if nomat == "OXT":
                nomr = nomr + "C"
                print(f"WARNING replacing {nomr[:-1]}:{line[23:27]} by {nomr}")
            ######################################################
            parms = aaLib.getParams(nomr, nomat)
            charges_list.append(parms.charg)
        return charges_list


def get_pdb_cmip_elements_canonical(input_pdb_filename: str, residue_library_path: str = None) -> List:
    if not residue_library_path:
        residue_library_path = str(Path(__file__).parent.joinpath("dat","aa.lib").resolve())

    aaLib = ResiduesDataLib(residue_library_path)
    print("{} residue/atom pairs loaded from {}".format(aaLib.nres, residue_library_path))

    with open(input_pdb_filename) as inPDB:
        elements_list = []
        residue_num = None
        for line in inPDB:
            line = line.rstrip()
            if not re.match('^ATOM', line) and not re.match('^HETATM', line):
                continue

            nomat = line[12:16]
            if re.match('^[1-9]', nomat):
                nomat = nomat[1:4] + nomat[:1]
            nomat = nomat.replace(' ', '')
            nomr = line[17:21].replace(' ', '')
            # WARNING: Temporal totally uninformed assumption by PA
            if nomr == "HIS":
                nomr = "HID"
                if residue_num != line[23:27]:
                    print(f"WARNING replacing HIS:{line[23:27]} by HID")
                    residue_num = line[23:27]
            # Thats not correct REVIEW this should be done for all the atoms in the residue
            # not just the oxigen
            if nomat == "OXT":
                nomr = nomr + "C"
                print(f"WARNING replacing {nomr[:-1]}:{line[23:27]} by {nomr}")
            ######################################################
            parms = aaLib.getParams(nomr, nomat)
            elements_list.append(parms.atType)
        return elements_list


def get_pdb_total_charge(pdb_file_path: str):
    # Biopython 1.9 does not capture charge of atoms in CMIP format
    # Should do it by hand
    total_charge = 0
    with open(pdb_file_path) as pdb_file:
        for line in pdb_file:
            if line[0:6].strip().upper() in ["ATOM", "HETATM"] and len(line) > 63:
                total_charge += float(line[55:63].strip())
    return total_charge


def probe_params_grid(probe_id: int = 0, readgrid: int = 2, pbfocus: int = 1, perfill: float = 0.6,
                      grid_int: Sequence[float] = (0.5, 0.5, 0.5)) -> Dict[str, str]:
    grid_dict = {}
    grid_dict[f"readgrid{probe_id}"] = f"{readgrid}"
    grid_dict[f"perfill{probe_id}"] = f"{perfill}"
    grid_dict['pbfocus'] = f"{pbfocus}"
    grid_dict['grid_int'] = f"INTX{probe_id}={grid_int[0]},INTY{probe_id}={grid_int[1]},INTZ{probe_id}={grid_int[2]}"

    return grid_dict


def params_grid(grid_type: str, readgrid: int = 0, perfill: float = 0.8,
                grid_int: Sequence[float] = (0.5, 0.5, 0.5),
                grid_dim: Sequence[float] = (64, 64, 64),
                grid_cen: Sequence[float] = (0.0, 0.0, 0.0)) -> Dict[str, str]:
    # grid_type older readgrid equivalences:
    #     2 proteina dist minima pecentatge, 4 distancia minima prot, 5 distancia al centre de masses
    #     1
    #     interaction = 0 , 3 explicita grid d'entrada
    #     cmip, titration, pbsolvation = 2, >3

    grid_dict = {}
    grid_dict[f"readgrid"] = f"{readgrid}"

    if grid_type in ['pb_interaction_energy', 'mip_pos', 'mip_neu', 'mip_neg', 'docking']:
        grid_dict['grid_cen'] = f"CENX={grid_cen[0]},CENY={grid_cen[1]},CENZ={grid_cen[2]}"
        grid_dict['grid_dim'] = f"DIMX={grid_dim[0]},DIMY={grid_dim[1]},DIMZ={grid_dim[2]}"
        grid_dict['grid_int'] = f"INTX={grid_int[0]},INTY={grid_int[1]},INTZ={grid_int[2]}"
    elif grid_type in ['solvation', 'titration']:
        grid_dict['perfill'] = f"{perfill}"
        grid_dict['grid_int'] = f"INTX={grid_int[0]},INTY={grid_int[1]},INTZ={grid_int[2]}"

    return grid_dict


def params_preset(execution_type: str) -> Dict[str, str]:
    params_dict = {}
    grid_dict = {}
    probe_grid_dict = {}
    execution_type = execution_type.strip()
    if execution_type == 'titration':
        grid_dict = params_grid(grid_type=execution_type, readgrid=2, perfill=0.8, grid_int=(0.5, 0.5, 0.5) )
        params_dict = {
            'title': 'Titration',
            'tipcalc': 1,
            'calcgrid': 1,
            'irest': 0,
            'orest': 0,
            'coorfmt': 2,
            'dields': 2,
            'titration': 1, 'inifoc': 2, 'cutfoc': -0.5, 'focus': 1, 'ninter': 10, 'clhost': 1, 'titcut': 20.,
            'titwat': 10, 'titip': 10, 'titim': 10
        }
    elif execution_type == 'mip_pos':
        grid_dict = params_grid(grid_type=execution_type, readgrid=2)
        params_dict = {
            'title': 'MIP positive probe',
            'tipcalc': 0,
            'calcgrid': 1,
            'irest': 0,
            'orest': 0,
            'coorfmt': 2,
            'dields':  2,
            'cubeoutput': 1,
            'fvdw': 0.8,
            'carmip': 1,
            'tipatmip' : "'OW'"
        }
    elif execution_type == 'mip_neu':
        grid_dict = params_grid(grid_type=execution_type, readgrid=2)
        params_dict = {
            'title': 'MIP neutral probe',
            'tipcalc': 0,
            'calcgrid': 1,
            'irest': 0,
            'orest': 0,
            'coorfmt': 2,
            'dields':  2,
            'cubeoutput': 1,
            'fvdw': 0.8,
            'carmip': 0,
            'tipatmip' : "'OW'"
        }
    elif execution_type == 'mip_neg':
        grid_dict = params_grid(grid_type=execution_type, readgrid=2)
        params_dict = {
            'title': 'MIP negative probe',
            'tipcalc': 0,
            'calcgrid': 1,
            'irest': 0,
            'orest': 0,
            'coorfmt': 2,
            'dields':  2,
            'cubeoutput': 1,
            'fvdw': 0.8,
            'carmip': -1,
            'tipatmip' : "'OW'"
        }
#TODO 'carmip': 1,
    # wat: tipcalc: 1 + titration: 'inifoc': 2, 'cutfoc': -0.5, 'focus': 1, 'ninter': 10,
    elif execution_type == 'solvation':
        grid_dict = params_grid(grid_type=execution_type, readgrid=2, perfill=0.2,
                                grid_int=(0.5, 0.5, 0.5))
        params_dict = {
            'title': 'Solvation & MEP',
            'tipcalc': 0,
            'calcgrid': 1,
            'irest': 0,
            'orest': 0,
            'coorfmt': 2,
            'cubeoutput': 1, 'vdw': 0,  'pbelec': 1,
            'novdwgrid': 1, 'solvenergy': 1, 'dielc': 1, 'dielsol': 80
        }

    elif execution_type == 'pb_interaction_energy':
        grid_dict = params_grid(grid_type=execution_type, readgrid= 2)
        probe_grid_dict = probe_params_grid(probe_id= 0, readgrid= 2, pbfocus= 1, perfill= 0.6,
                                            grid_int=(1.5, 1.5, 1.5))

        # TODO Check for external box file or parameters
        params_dict = {
            'title': 'Docking Interaction energy calculation. PB electrostatics',
            'tipcalc': 3,
            'calcgrid': 1,
            'irest': 0,
            'orest': 0,
            'coorfmt': 2,
            'fvdw': 0.8, 'pbelec': 1, 'pbinic': 2, 'wgp': 0, 'ebyatom': 1
        }

    elif execution_type == 'docking':
        grid_dict = params_grid(grid_type=execution_type, readgrid= 2)

        params_dict = {
            'title': 'Docking Mehler Solmajer dielectric',
            'tipcalc': 2,
            'calcgrid': 1,
            'irest': 0,
            'orest': 1,
            'coorfmt': 2,
            'fvdw': 0.8, 'dields': 2, 'focus': 1, 'cutfoc': 100,
            'tiprot': 5, 'inifoc': 5, 'ninter': 20,
            'clhost': 1, 'minout': 50, 'splitpdb': 0
        }

    elif execution_type == 'docking_rst':
        params_dict = {
            'title': 'Docking from restart file',
            'readgrid': 0,
            'tipcalc': 2,
            'calcgrid': 1,
            'irest': 2,
            'orest': 1,
            'coorfmt': 2,
            'fvdw': 0.8, 'dields': 2, 'focus': 1, 'cutfoc': 100,
            'tiprot': 5, 'inifoc': 5, 'ninter': 20,
            'clhost': 1, 'minout': 50, 'splitpdb': 0, 'cutelec': 10.0
        }
    elif execution_type == 'check_only':
        params_dict = {
            'title': 'Check_only dry run of CMIP',
            'CHECKONLY': 1,
            'readgrid': 2,
            'calcgrid': 1,
            'tipcalc': 0,
            'irest': 0,
            'ebyatom': 1,
            'coorfmt': 2,
            'fvdw': 0.8
        }

    return {**params_dict, **grid_dict, **probe_grid_dict}


def read_params_file(input_params_path: str) -> Dict[str, str]:
    params_dict = {}
    with open(input_params_path) as input_params_file:
        params_dict['title']: input_params_file.readline()
        for line in input_params_file:
            line = line.replace(' ', '')
            if line.startswith('&'): continue
            param_list = line.split(',')
            for param in param_list:
                param_key, param_value = param.split("=")

                # Grid Values
                if len(param_key) > 3 and param_key[:3].startswith('INT'):
                    if params_dict.get('grid_int'):
                        params_dict['grid_int'] += f",{param_key}={param_value}"
                    else:
                        params_dict['grid_int'] = f"{param_key}={param_value}"
                elif len(param_key) > 3 and param_key[:3].startswith('CEN'):
                    if params_dict.get('grid_cen'):
                        params_dict['grid_cen'] += f",{param_key}={param_value}"
                    else:
                        params_dict['grid_cen'] = f"{param_key}={param_value}"
                elif len(param_key) > 3 and param_key[:3].startswith('DIM'):
                    if params_dict.get('grid_dim'):
                        params_dict['grid_dim'] += f",{param_key}={param_value}"
                    else:
                        params_dict['grid_dim'] = f"{param_key}={param_value}"
                # Rest of parameters
                else:
                    params_dict[param_key] = param_value
    return params_dict


def write_params_file(output_params_path: str, params_dict: Mapping[str, str]) -> str:
    with open(output_params_path, 'w') as output_params_file:
        output_params_file.write(f"{params_dict.pop('title', 'Untitled')}\n")
        output_params_file.write(f"&cntrl\n")
        for params_key, params_value in params_dict.items():
            if params_key in ['grid_int', 'grid_cen', 'grid_dim', 'grid_int0', 'grid_cen0', 'grid_dim0']:
                output_params_file.write(f" {params_value}\n")
            else:
                output_params_file.write(f" {params_key} = {params_value}\n")
        output_params_file.write(f"&end\n")
    return output_params_path


def create_params_file(output_params_path: str, input_params_path: str = None,
                       params_preset_dict: Mapping = None, params_properties_dict: Mapping = None) -> str:
    """ Gets a params dictionary and a presset and returns the path of the created params file for cmip.

    Args:


    Returns:
        str: params file path.
    """
    params_dict = {}

    if params_preset_dict:
        for k, v in params_preset_dict.items():
            params_dict[k] = v
    if input_params_path:
        input_params_dict = read_params_file(input_params_path)
        for k, v in input_params_dict.items():
            params_dict[k] = v
    if params_properties_dict:
        for k, v in params_properties_dict.items():
            params_dict[k] = v

    return write_params_file(output_params_path, params_dict)

def mark_residues(residue_list: Sequence[str], input_cmip_pdb_path: str, output_cmip_pdb_path: str, out_log: logging.Logger = None, global_log: logging.Logger = None) -> None:
    """Marks using an "X" before the atom type all the residues in *residue_list* and writes the result in *output_cmip_pdb_path*.

        Args:
            residue_list (list): Residue list in the format "Chain:Resnum" (no spaces between the elements) separated by commas. If empty or none all residues will be marked.
            local_log (:obj:`logging.Logger`): local log object.
            global_log (:obj:`logging.Logger`): global log object.
        """
    if not residue_list:
        fu.log(f"Empty residue_list all residues will be marked", out_log, global_log)
    else:
        fu.log(f"Residue list: {residue_list}", out_log, global_log)

    with open(input_cmip_pdb_path) as pdb_file_in, open(output_cmip_pdb_path, 'w') as pdb_file_out:
        residue_set_used = set()

        res_counter = 0
        for line in pdb_file_in:
            if _is_atom(line):
                residue_code = _get_residue_code(line)
                used_residue = _get_residue_code_in_list(residue_code, residue_list)
                if not residue_list or used_residue:
                    res_counter += 1
                    residue_set_used.add(used_residue)
                    line = _mark_pdb_atom(line)
            pdb_file_out.write(line)
        fu.log(f"{res_counter} residues have been marked", out_log, global_log)

        if residue_list:
            unused_residues = set(residue_list) - residue_set_used
            if unused_residues:
                fu.log(f"The following residues where present in the residue_list and have not been marked: {unused_residues}", out_log, global_log)

def _mark_pdb_atom(line: str) -> str:
    line = list(line)
    line.insert(64, 'X')
    return ''.join(line)

def _get_residue_code_in_list(input_residue, residue_list):
    if not residue_list:
        return None
    for res_code in residue_list:
        if input_residue == res_code:
            return res_code
        chain_rescode, resnum_rescode = res_code.split(":")
        chain_input, resnum_input = input_residue.split(":")
        if not chain_rescode:
            if resnum_rescode == resnum_input:
                return res_code
        if not resnum_rescode:
            if chain_rescode == chain_input:
                return res_code
    return None


def _get_residue_code(line: str) -> str:
    return _get_chain(line)+":"+_get_resnum(line)

def _get_chain(line: str) -> str:
    return line[21].upper()

def _get_resnum(line: str) -> str:
    return line[22:27].strip()

def _is_atom(line: str) -> str:
    return line[0:6].strip().upper() in ["ATOM", "HETATM"]



""" Representation functions for package biobb_cmip.cmip """
from biobb_cmip.cmip.common import get_grid
from pathlib import Path
from MDAnalysis.lib.util import inverse_aa_codes
from typing import List, Dict, Mapping, Union, Set, Sequence, Tuple, List


def get_energies_byat(cmip_energies_byat_out: Union[str, Path], cutoff: float = 100.0) -> Tuple[List[str], Dict[str, List[float]]]:

    with open(cmip_energies_byat_out, 'r') as energies_file:
        atom_list = []
        energy_dict = {"ES": [], "VDW": [], "ES&VDW": []}
        for line in energies_file:
            atom_list.append(line[6:12].strip())
            vdw = float(line[42:53]) if float(line[42:53]) < cutoff else 0.0
            es = float(line[57:68]) if float(line[57:68]) < cutoff else 0.0
            both = float(line[72:83]) if float(line[72:83]) < cutoff else 0.0

            energy_dict["ES"].append(es)
            energy_dict["VDW"].append(vdw)
            energy_dict["ES&VDW"].append(both)

    return atom_list, energy_dict


def get_energies_byres(cmip_energies_byat_out: Union[str, Path], cutoff: float = 100.0) -> Tuple[List[str], Dict[str, List[float]]]:
    residues = []
    energy_dict = {"ES": [], "VDW": [], "ES&VDW": []}
    with open(cmip_energies_byat_out, 'r') as energies_file:
        for line in energies_file:
            chain = line[21:22].strip()
            residue_id = line[22:28].strip()
#            residue_id = str(int(residue_id) + 697)
#            if (int(residue_id) > 746):
#                residue_id = str(int(residue_id) + 9)
#            if (int(residue_id) > 749):
#                residue_id = str(int(residue_id) + 4)
#            if (int(residue_id) > 867):
#                residue_id = str(int(residue_id) + 8)
#            if (int(residue_id) > 987):
#                residue_id = str(int(residue_id) + 7)
#            if (int(residue_id) > 1004):
#                residue_id = str(int(residue_id) + 5)
            resname = inverse_aa_codes.get(line[17:21].strip().upper(), "X")
            residue = resname+' '+chain+residue_id
            vdw = float(line[42:53]) if float(line[42:53]) < cutoff else 0.0
            es = float(line[57:68]) if float(line[57:68]) < cutoff else 0.0
            both = float(line[72:83]) if float(line[72:83]) < cutoff else 0.0

            if residue in residues:
                index = residues.index(residue)
                energy_dict["ES"][index] += es
                energy_dict["VDW"][index] += vdw
                energy_dict["ES&VDW"][index] += both
            else:
                residues.append(residue)
                #residues.append(int(residue_id)+696)
                energy_dict["ES"].append(es)
                energy_dict["VDW"].append(vdw)
                energy_dict["ES&VDW"].append(both)

        return residues, energy_dict


def create_box_representation(cmip_log_path: Union[str, Path], cmip_pdb_path: Union[str, Path]) -> Tuple[str, List[List[str]]]:
    return _create_box_representation_file(cmip_log_path, cmip_pdb_path), _get_atom_pair()


def _create_box_representation_file(cmip_log_path: Union[str, Path], cmip_pdb_path: Union[str, Path]) -> str:
    vertex_list = _get_vertex_list(cmip_log_path)

    cmip_pdb_path = Path(cmip_pdb_path).resolve()
    boxed_pdb_path = cmip_pdb_path.parent.joinpath("boxed_"+str(cmip_pdb_path.name))
    with open(cmip_pdb_path) as cmip_pdb_file:
        pdb_lines = cmip_pdb_file.readlines()
        if pdb_lines[-1].strip().upper() == "END":
            pdb_lines = pdb_lines[:-1]
    with open(boxed_pdb_path, 'w') as boxed_pdb_file:
        for pdb_line in pdb_lines:
            boxed_pdb_file.write(pdb_line)
        for i, v in enumerate(vertex_list):
            boxed_pdb_file.write('HETATM10000 ZN' + str(i) + '   ZN Z9999    ' + v + '  1.00 50.00          ZN\n')
        boxed_pdb_file.write("END")
    return str(boxed_pdb_path)


def _get_vertex_list(cmip_log_path: Union[str, Path]) -> List[str]:
    origin, size, _ = get_grid(cmip_log_path)
    return [
        _pdb_coord_formatter(origin[0]) + _pdb_coord_formatter(origin[1]) + _pdb_coord_formatter(origin[2]),
        _pdb_coord_formatter(origin[0] + size[0]) + _pdb_coord_formatter(origin[1]) + _pdb_coord_formatter(origin[2]),
        _pdb_coord_formatter(origin[0]) + _pdb_coord_formatter(origin[1] + size[1]) + _pdb_coord_formatter(origin[2]),
        _pdb_coord_formatter(origin[0]) + _pdb_coord_formatter(origin[1]) + _pdb_coord_formatter(origin[2] + size[2]),
        _pdb_coord_formatter(origin[0] + size[0]) + _pdb_coord_formatter(origin[1] + size[1]) + _pdb_coord_formatter(origin[2]),
        _pdb_coord_formatter(origin[0] + size[0]) + _pdb_coord_formatter(origin[1]) + _pdb_coord_formatter(origin[2] + size[2]),
        _pdb_coord_formatter(origin[0]) + _pdb_coord_formatter(origin[1] + size[1]) + _pdb_coord_formatter(origin[2] + size[2]),
        _pdb_coord_formatter(origin[0] + size[0]) + _pdb_coord_formatter(origin[1] + size[1]) + _pdb_coord_formatter(origin[2] + size[2]),
    ]

def _pdb_coord_formatter(coordinate: float) -> str:
    return str(round(coordinate, 3)).rjust(8)


def _get_atom_pair() -> List[List[str]]:
    return [[ "9999:Z.ZN0", "9999:Z.ZN1" ],
            [ "9999:Z.ZN0", "9999:Z.ZN2" ],
            [ "9999:Z.ZN0", "9999:Z.ZN3" ],
            [ "9999:Z.ZN1", "9999:Z.ZN4" ],
            [ "9999:Z.ZN1", "9999:Z.ZN5" ],

            [ "9999:Z.ZN2", "9999:Z.ZN4" ],
            [ "9999:Z.ZN2", "9999:Z.ZN6" ],

            [ "9999:Z.ZN3", "9999:Z.ZN5" ],
            [ "9999:Z.ZN3", "9999:Z.ZN6" ],

            [ "9999:Z.ZN4", "9999:Z.ZN7" ],
            [ "9999:Z.ZN5", "9999:Z.ZN7" ],
            [ "9999:Z.ZN6", "9999:Z.ZN7" ]]

# AUTOMATIC OUTER GRID / titleGrid0=Automatic Outer Grid
#
# INT:
#     spacing=   1.50000000     ,   1.50000000     ,   1.50000000
#     Grid units:      1.500   1.500   1.500
# CEN:
#     center=   71.6100006     ,   67.7550049     ,   56.3150024
#     Grid center:    71.610  67.755  56.315
# DIM:
#     dim=          64 ,          92 ,          72
#     Grid density:    64   92   72
#
# To create graphic representations:
# Size:
#     Grid Size:      96.000 138.000 108.000  90.000  90.000  90.000
#     size=   96.0000000     ,   138.000000     ,   108.000000
#
# Origin:
#     Grid origin:    23.610  -1.245   2.315
#     origin=   23.6100006     ,  -1.24499512     ,   2.31500244








import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="biobb_cmip",
    version="3.9.0",
    author="Biobb developers",
    author_email="pau.andrio@bsc.es",
    description="biobb_cmip is the Biobb module collection to compute classical molecular interaction potentials.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="Bioinformatics Workflows BioExcel Compatibility",
    url="https://github.com/bioexcel/biobb_cmip",
    project_urls={
        "Documentation": "http://biobb_cmip.readthedocs.io/en/latest/",
        "Bioexcel": "https://bioexcel.eu/"
    },
    packages=setuptools.find_packages(exclude=['docs', 'test']),
    include_package_data=True,
    install_requires=[
        'biobb_common==3.9.0',
        'mdanalysis==2.0.0',
        'biobb_structure_checking==3.12.1'
    ],
    python_requires='>=3.7',
    entry_points={
        "console_scripts": [
            "cmip = biobb_cmip.cmip.cmip:main",
            "titration = biobb_cmip.cmip.titration:main",
            "prepare_structure = biobb_cmip.cmip.prepare_structure:main",
            "prepare_pdb = biobb_cmip.cmip.prepare_pdb:main",
            "ignore_residues = biobb_cmip.cmip.ignore_residues:main"
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
    ],
)



#!/usr/bin/env python3

"""Module containing the Cmip class and the command line interface."""
import os
import json
import argparse
import shutil
from pathlib import Path
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger
from biobb_cmip.cmip.common import create_params_file
from biobb_cmip.cmip.common import params_preset
from biobb_cmip.cmip.common import get_grid


class Cmip(BiobbObject):
    """
    | biobb_cmip Titration
    | Wrapper class for the CMIP cmip module.
    | The CMIP cmip module. CMIP cmip module compute classical molecular interaction potentials.

    Args:
        input_pdb_path (str): Path to the input PDB file. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_cmip/master/biobb_cmip/test/data/cmip/1kim_h.pdb>`_. Accepted formats: pdb (edam:format_1476).
        input_probe_pdb_path (str) (Optional): Path to the input probe file in PDB format. File type: input. Accepted formats: pdb (edam:format_1476).
        output_pdb_path (str) (Optional): Path to the output PDB file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_cmip/master/biobb_cmip/test/reference/cmip/1kim_neutral.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_grd_path (str) (Optional): Path to the output grid file in GRD format. File type: output. Accepted formats: grd (edam:format_2330).
        output_cube_path (str) (Optional): Path to the output grid file in cube format. File type: output. Accepted formats: cube (edam:format_2330).
        output_rst_path (str) (Optional): Path to the output restart file. File type: output. Accepted formats: txt (edam:format_2330).
        input_rst_path (str) (Optional): Path to the input restart file. File type: output. Accepted formats: txt (edam:format_2330).
        output_byat_path (str) (Optional): Path to the output atom by atom energy file. File type: output. Accepted formats: txt (edam:format_2330), out (edam:format_2330).
        output_log_path (str) (Optional): Path to the output CMIP log file LOG. File type: output. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/ref_cmip.log>`_. Accepted formats: log (edam:format_2330).
        input_vdw_params_path (str) (Optional): Path to the CMIP input Van der Waals force parameters, if not provided the CMIP conda installation one is used ("$CONDA_PREFIX/share/cmip/dat/vdwprm"). File type: input. Accepted formats: txt (edam:format_2330).
        input_params_path (str) (Optional): Path to the CMIP input parameters file. File type: input. Accepted formats: txt (edam:format_2330).
        output_json_box_path (str) (Optional): Path to the output CMIP box in JSON format. File type: output. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/ref_box.json>`_. Accepted formats: json (edam:format_3464).
        output_json_external_box_path (str) (Optional): Path to the output external CMIP box in JSON format. File type: output. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/ref_box.json>`_. Accepted formats: json (edam:format_3464).
        input_json_box_path (str) (Optional): Path to the input CMIP box in JSON format. File type: input. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/ref_box.json>`_. Accepted formats: json (edam:format_3464).
        input_json_external_box_path (str) (Optional): Path to the input CMIP box in JSON format. File type: input. `Sample file <https://github.com/bioexcel/biobb_cmip/raw/master/biobb_cmip/test/reference/cmip/ref_box.json>`_. Accepted formats: json (edam:format_3464).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **execution_type** (*str*) - ("mip_pos") Default options for the params file, each one creates a different params file. Values: check_only (Dry Run of CMIP), mip_pos (MIP O+  Mehler Solmajer dielectric), mip_neg (MIP O-  Mehler Solmajer dielectric), mip_neu (MIP Oxygen Mehler Solmajer dielectric), solvation (Solvation & MEP), pb_interaction_energy (Docking Interaction energy calculation. PB electrostatics), docking (Docking Mehler Solmajer dielectric), docking_rst (Docking from restart file).
            * **params** (*dict*) - ({}) CMIP options specification.
            * **binary_path** (*str*) - ("cmip") Path to the CMIP cmip executable binary.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **container_path** (*str*) - (None)  Path to the binary executable of your container.
            * **container_image** (*str*) - ("cmip/cmip:latest") Container Image identifier.
            * **container_volume_path** (*str*) - ("/data") Path to an internal directory in the container.
            * **container_working_dir** (*str*) - (None) Path to the internal CWD in the container.
            * **container_user_id** (*str*) - (None) User number id to be mapped inside the container.
            * **container_shell_path** (*str*) - ("/bin/bash") Path to the binary executable of the container shell.


    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_cmip.cmip.cmip import cmip
            prop = { 'binary_path': 'cmip' }
            cmip(input_pdb_path='/path/to/myStructure.pdb',
                      output_pdb_path='/path/to/newStructure.pdb',
                      output_log_path='/path/to/newStructureLog.log',
                      properties=prop)

    Info:
        * wrapped_software:
            * name: CMIP cmip
            * version: 2.7.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(self, input_pdb_path: str, input_probe_pdb_path: str = None, output_pdb_path: str = None,
                 output_grd_path: str = None, output_cube_path: str = None, output_rst_path: str = None,
                 input_rst_path: str = None, output_byat_path: str = None, output_log_path: str = None,
                 input_vdw_params_path: str = None, input_params_path: str = None, output_json_box_path: str = None,
                 output_json_external_box_path: str = None, input_json_box_path: str = None,
                 input_json_external_box_path: str = None, properties: dict = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_pdb_path": input_pdb_path, "input_probe_pdb_path": input_probe_pdb_path,
                   "input_vdw_params_path": input_vdw_params_path, "input_params_path": input_params_path,
                   "input_json_box_path": input_json_box_path,
                   "input_json_external_box_path": input_json_external_box_path,
                   "input_rst_path": input_rst_path},
            "out": {"output_pdb_path": output_pdb_path, "output_grd_path": output_grd_path,
                    "output_cube_path": output_cube_path, "output_rst_path": output_rst_path,
                    "output_byat_path": output_byat_path, "output_log_path": output_log_path,
                    "output_json_box_path": output_json_box_path,
                    "output_json_external_box_path": output_json_external_box_path}
        }

        # Properties specific for BB
        self.binary_path = properties.get('binary_path', 'cmip')
        self.execution_type = properties.get('execution_type', 'mip_pos')
        self.params = {k: str(v) for k, v in properties.get('params', dict()).items()}

        if not self.io_dict['in'].get('input_vdw_params_path'):
            self.io_dict['in']['input_vdw_params_path'] = f"{os.environ.get('CONDA_PREFIX')}/share/cmip/dat/vdwprm"
        self.io_dict['in']['combined_params_path'] = properties.get('combined_params_path', 'params')

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`Cmip <cmip.cmip.Cmip>` object."""
        tmp_files = []

        # Setup Biobb
        if self.check_restart():
            return 0

        # Check if output_pdb_path ends with ".pdb" and does not contain underscores
        if self.io_dict['out']['output_pdb_path']:
            if (not self.io_dict['out']['output_pdb_path'].endswith('.pdb')) or \
               ("_" in str(Path(self.io_dict['out']['output_pdb_path']).name)):
                fu.log(f"ERROR: output_pdb_path ({self.io_dict['out']['output_pdb_path']}) "
                       f"name must end in .pdb and not contain underscores", self.out_log, self.global_log)
                raise ValueError(f"ERROR: output_pdb_path ({self.io_dict['out']['output_pdb_path']})"
                                 f"name must end in .pdb and not contain underscores")

        params_preset_dict = params_preset(execution_type=self.execution_type)
        if self.io_dict['in']["input_json_external_box_path"]:
            params_preset_dict["readgrid0"] = 0
            origin, size, grid_params = get_grid(self.io_dict['in']["input_json_external_box_path"])
            params_preset_dict['grid_int0'] = \
                f"INTX0={grid_params['INT'][0]},INTY0={grid_params['INT'][1]},INTZ0={grid_params['INT'][2]}"
            params_preset_dict['grid_cen0'] = \
                f"CENX0={grid_params['CEN'][0]},CENY0={grid_params['CEN'][1]},CENZ0={grid_params['CEN'][2]}"
            params_preset_dict['grid_dim0'] = \
                f"DIMX0={grid_params['DIM'][0]},DIMY0={grid_params['DIM'][1]},DIMZ0={grid_params['DIM'][2]}"

        if self.io_dict['in']["input_json_box_path"]:
            params_preset_dict["readgrid"] = 0
            origin, size, grid_params = get_grid(self.io_dict['in']["input_json_box_path"])
            params_preset_dict['grid_int'] = \
                f"INTX={grid_params['INT'][0]},INTY={grid_params['INT'][1]},INTZ={grid_params['INT'][2]}"
            params_preset_dict['grid_cen'] = \
                f"CENX={grid_params['CEN'][0]},CENY={grid_params['CEN'][1]},CENZ={grid_params['CEN'][2]}"
            params_preset_dict['grid_dim'] = \
                f"DIMX={grid_params['DIM'][0]},DIMY={grid_params['DIM'][1]},DIMZ={grid_params['DIM'][2]}"

        if self.io_dict['out']['output_json_box_path'] or self.io_dict['out']['output_json_external_box_path']:
            params_preset_dict['WRITELOG'] = 1
            key_value_log_dir = fu.create_unique_dir()
            self.io_dict['out']['key_value_log_path'] = str(Path(key_value_log_dir).joinpath("key_value_cmip_log.log"))
            self.tmp_files.append(key_value_log_dir)

        # Restart OUT
        if self.io_dict["out"].get("output_rst_path"):
            params_preset_dict['FULLRST'] = 1
            params_preset_dict['OREST'] = 1

        # Restart IN
        if self.io_dict['in']["input_rst_path"]:
            params_preset_dict['IREST'] = 2
            if not self.io_dict["out"].get("output_rst_path"):
                self.io_dict["out"]["output_rst_path"] = fu.create_unique_file_path()
            shutil.copy2(self.io_dict['in']["input_rst_path"], self.io_dict["out"]["output_rst_path"])

        else:
            params_preset_dict['IREST'] = 0

        combined_params_dir = fu.create_unique_dir()
        self.io_dict['in']['combined_params_path'] = create_params_file(
            output_params_path=str(Path(combined_params_dir).joinpath(self.io_dict['in']['combined_params_path'])),
            input_params_path=self.io_dict['in'].get('input_params_path'),
            params_preset_dict=params_preset_dict,
            params_properties_dict=self.params)

        self.stage_files()

        self.cmd = [self.binary_path,
                    '-i', self.stage_io_dict['in']['combined_params_path'],
                    '-vdw', self.stage_io_dict['in']['input_vdw_params_path'],
                    '-hs', self.stage_io_dict['in']['input_pdb_path']]

        if self.stage_io_dict["in"].get("input_probe_pdb_path") and Path(
                self.io_dict["in"].get("input_probe_pdb_path")).exists():
            self.cmd.append('-pr')
            self.cmd.append(self.stage_io_dict["in"].get("input_probe_pdb_path"))

        if self.stage_io_dict["out"].get("output_pdb_path"):
            self.cmd.append('-outpdb')
            self.cmd.append(self.stage_io_dict['out']['output_pdb_path'])

        if self.stage_io_dict["out"].get("output_grd_path"):
            self.cmd.append('-grdout')
            self.cmd.append(self.stage_io_dict["out"]["output_grd_path"])

        if self.stage_io_dict["out"].get("output_cube_path"):
            self.cmd.append('-cube')
            self.cmd.append(self.stage_io_dict["out"]["output_cube_path"])

        if self.stage_io_dict["out"].get("output_rst_path"):
            self.cmd.append('-rst')
            self.cmd.append(self.stage_io_dict["out"]["output_rst_path"])

        if self.stage_io_dict["out"].get("output_byat_path"):
            self.cmd.append('-byat')
            self.cmd.append(self.stage_io_dict["out"]["output_byat_path"])

        if self.stage_io_dict["out"].get("output_log_path"):
            self.cmd.append('-o')
            self.cmd.append(self.stage_io_dict["out"]["output_log_path"])

        if self.stage_io_dict['out'].get('output_json_box_path') or self.stage_io_dict['out'].get('output_json_external_box_path'):
            self.cmd.append('-l')
            self.cmd.append(self.stage_io_dict["out"]["key_value_log_path"])

        # Run Biobb block
        self.run_biobb()

        # CMIP removes or adds a .pdb extension from pdb output name
        # manual copy_to_host or unstage
        if self.io_dict['out'].get('output_pdb_path'):
            output_pdb_path = str(Path(self.stage_io_dict["unique_dir"]).joinpath(Path(self.io_dict['out'].get('output_pdb_path')).name))
            if Path(output_pdb_path[:-4]).exists():
                shutil.move(output_pdb_path[:-4], self.io_dict['out'].get('output_pdb_path'))
            elif Path(output_pdb_path + ".pdb").exists():
                shutil.move(output_pdb_path + ".pdb", self.io_dict['out'].get('output_pdb_path'))

        # Replace "ATOMTM" tag for "ATOM  "
        output_pdb_path = self.io_dict['out'].get('output_pdb_path')
        if output_pdb_path:
            with open(output_pdb_path) as pdb_file:
                list_pdb_lines = pdb_file.readlines()
            with open(output_pdb_path, 'w') as pdb_file:
                for line in list_pdb_lines:
                    pdb_file.write(line.replace('ATOMTM', 'ATOM  '))

        # Create json_box_path file from CMIP log file
        if self.io_dict['out'].get('output_json_box_path'):
            origin, size, grid_params = get_grid(self.stage_io_dict["out"]["output_log_path"])
            grid_params['DIM'] = (int(grid_params['DIM'][0]),
                                  int(grid_params['DIM'][1]),
                                  int(grid_params['DIM'][2]))
            size_dict = {'x': round(grid_params['DIM'][0] * grid_params['INT'][0], 3),
                         'y': round(grid_params['DIM'][1] * grid_params['INT'][1], 3),
                         'z': round(grid_params['DIM'][2] * grid_params['INT'][2], 3)}
            origin_dict = {'x': round(grid_params['CEN'][0] - size_dict['x'] / 2, 3),
                           'y': round(grid_params['CEN'][1] - size_dict['y'] / 2, 3),
                           'z': round(grid_params['CEN'][0] - size_dict['z'] / 2, 3)}
            grid_dict = {'origin': origin_dict,
                         'size': size_dict,
                         'params': grid_params}
            with open(self.io_dict['out'].get('output_json_box_path'), 'w') as json_file:
                json_file.write(json.dumps(grid_dict, indent=4))

        # Create external_json_box_path file from CMIP log file
        if self.io_dict['out'].get('output_json_external_box_path'):
            origin, size, grid_params = get_grid(self.stage_io_dict["out"]["output_log_path"], True)
            grid_params['DIM'] = (int(grid_params['DIM'][0]),
                                  int(grid_params['DIM'][1]),
                                  int(grid_params['DIM'][2]))
            size_dict = {'x': round(grid_params['DIM'][0] * grid_params['INT'][0], 3),
                         'y': round(grid_params['DIM'][1] * grid_params['INT'][1], 3),
                         'z': round(grid_params['DIM'][2] * grid_params['INT'][2], 3)}
            origin_dict = {'x': round(grid_params['CEN'][0] - size_dict['x'] / 2, 3),
                           'y': round(grid_params['CEN'][1] - size_dict['y'] / 2, 3),
                           'z': round(grid_params['CEN'][0] - size_dict['z'] / 2, 3)}
            grid_dict = {'origin': origin_dict,
                         'size': size_dict,
                         'params': grid_params}
            with open(self.io_dict['out'].get('output_json_external_box_path'), 'w') as json_file:
                json_file.write(json.dumps(grid_dict, indent=4))

        # Copy files to host
        self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([self.stage_io_dict.get("unique_dir"), combined_params_dir])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def cmip(input_pdb_path: str, input_probe_pdb_path: str = None, output_pdb_path: str = None,
         output_grd_path: str = None, output_cube_path: str = None, output_rst_path: str = None,
         output_byat_path: str = None, output_log_path: str = None,
         input_vdw_params_path: str = None, input_params_path: str = None, output_json_box_path: str = None,
         output_json_external_box_path: str = None, input_json_box_path: str = None,
         input_json_external_box_path: str = None, properties: dict = None, **kwargs) -> int:
    """Create :class:`Cmip <cmip.cmip.Cmip>` class and
    execute the :meth:`launch() <cmip.cmip.Cmip.launch>` method."""

    return Cmip(input_pdb_path=input_pdb_path, input_probe_pdb_path=input_probe_pdb_path,
                output_pdb_path=output_pdb_path, output_grd_path=output_grd_path, output_cube_path=output_cube_path,
                output_rst_path=output_rst_path, output_byat_path=output_byat_path, output_log_path=output_log_path,
                input_vdw_params_path=input_vdw_params_path, input_params_path=input_params_path,
                output_json_box_path=output_json_box_path, output_json_external_box_path=output_json_external_box_path,
                input_json_box_path=input_json_box_path, input_json_external_box_path=input_json_external_box_path,
                properties=properties, **kwargs).launch()


def main():
    parser = argparse.ArgumentParser(description="Wrapper of the CMIP cmip module.",
                                     formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('-c', '--config', required=False, help="This file can be a YAML file, JSON file or JSON string")

    # Specific args of each building block
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pdb_path', required=True)
    parser.add_argument('--input_probe_pdb_path', required=False)
    parser.add_argument('--output_pdb_path', required=False)
    parser.add_argument('--output_grd_path', required=False)
    parser.add_argument('--output_cube_path', required=False)
    parser.add_argument('--output_rst_path', required=False)
    parser.add_argument('--output_byat_path', required=False)
    parser.add_argument('--output_log_path', required=False)
    parser.add_argument('--input_vdw_params_path', required=False)
    parser.add_argument('--input_params_path', required=False)
    parser.add_argument('--output_json_box_path', required=False)
    parser.add_argument('--output_json_external_box_path', required=False)
    parser.add_argument('--input_json_box_path', required=False)
    parser.add_argument('--input_json_external_box_path', required=False)

    args = parser.parse_args()
    config = args.config if args.config else None
    properties = settings.ConfReader(config=config).get_prop_dic()

    # Specific call of each building block
    cmip(input_pdb_path=args.input_pdb_path, input_probe_pdb_path=args.input_probe_pdb_path,
         output_pdb_path=args.output_pdb_path, output_grd_path=args.output_grd_path,
         output_cube_path=args.output_cube_path, output_rst_path=args.output_rst_path,
         output_byat_path=args.output_byat_path, output_log_path=args.output_log_path,
         input_vdw_params_path=args.input_vdw_params_path, input_params_path=args.input_params_path,
         output_json_box_path=args.output_json_box_path,
         output_json_external_box_path=args.output_json_external_box_path, input_json_box_path=args.input_json_box_path,
         input_json_external_box_path=args.input_json_external_box_path, properties=properties)


if __name__ == '__main__':
    main()
